#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

//#define DEBUG_BITS
//#define DEBUG_SHOW_CLIP
//#define DEBUG_SHOW_ALL_ORIENTS
//#define DEBUG_SHOW_PARTS
//#define DEBUG_RECURSE_CHECK
//#define DEBUG_RECURSE
//#define DEBUG_SOLUTION_ALL
//#define DEBUG_SOLUTION_DETAILS
#define DEBUG_OPENINGS

typedef unsigned long      u32;
typedef unsigned long long u64;
typedef u32                bool;
#define false 0
#define true  1


/*********************/
/* Puzzle properties */
/*********************/

#define PARTS        12
#define ORIENTATIONS 184
#define PART_ORIENTS 24

#define DX 3  // dimension x
#define DY 4
#define DZ 5

// correction factors when rotating axis a into axis b
#define DX2Y (DX>DY? DX-DY : 0)
#define DY2X (DY>DX? DY-DX : 0)
#define DX2Z (DX>DZ? DX-DZ : 0)
#define DZ2X (DZ>DX? DZ-DX : 0)
#define DY2Z (DY>DZ? DY-DZ : 0)
#define DZ2Y (DZ>DY? DZ-DY : 0)

#define MASK(n)          (((u64)1) << (n))
#define XYZOFFSET(x,y,z) ((x)+DX*(y)+DX*DY*(z))
#define XYZMASK(x,y,z)   MASK(XYZOFFSET(x,y,z))
#define XYZTEST(x,y,z,d) ((d) & XYZMASK(x,y,z))

/* neighbors in positive x,y,z directions */
#define NEIGHBORS (XYZMASK(1,0,0) | XYZMASK(0,1,0) | XYZMASK(0,0,1))


const char *p0 []={"xx",
                   "x ",
                   "xx",NULL,NULL};
const char *p1 []={" xx",
                   " x ",
                   "xx ",NULL,NULL};
const char *p2 []={" x",
                   " x",
                   " x",
                   "xx",NULL,NULL};
const char *p3 []={"x ",
                   "xx",
                   "x ",
                   "x ",NULL,NULL};
const char *p4 []={"x ",
                   "x ",NULL,
                   //
                   "x ",
                   "xx",NULL,NULL};
const char *p5 []={"xxx",
                   " x ",
                   " x ",NULL,NULL};
const char *p6 []={" x",
                   "xx",
                   "xx",NULL,NULL};
const char *p7 []={"  x",
                   " xx",
                   "xx ",NULL,NULL};
const char *p8 []={"  ",
                   "x ",NULL,
                   //
                   " x",
                   "xx",NULL,NULL};
const char *p9 []={"x  ",
                   "xxx",
                   " x ",NULL,NULL};
const char *p10[]={" x ",
                   "xxx",
                   " x ",NULL,NULL};
const char *p11[]={"x ",
                   "xx",
                   "x ",
                   "xx",NULL,NULL};

const char **partConsts[]={p0,p1,p5,p7,p8,p2,p3,p11,p4,p6,p9,p10,NULL};


/*************/
/* Bit stuff */
/*************/

/* Print a bitmask as a part / board */
void bmaskPrint(u64 data) {
  int x,y,z;
  for(y=0; y<DY; y++) {
    for(z=0; z<DZ; z++) {
      for(x=0; x<DX; x++)
        fputc(XYZTEST(x,y,z,data) ? 'X' : '.', stdout);
      fputc(' ',stdout);
    }
    puts("");
  }
  puts("");
}

/* Initialize part bitmask from string array.
   ' ' is empty, else nonempty. A string is a line.
   Terminate planes with NULL, part with NULL,NULL. */
u64 bmaskInit(const char *init[]) {
  u64 data=0;
  int line=0;

  int x,y,z;
  for(z=0; init[line]; z++, line++)
    for(y=0; init[line]; y++, line++)
      for(x=0; init[line][x]; x++)
        if(init[line][x]!=' ')
          data |= XYZMASK(x,y,z);
  return data;
}

/* Find first clear bit in 64-bit word */
inline int bmaskFirstClear64(u32 dlow, u32 dhigh) {
  register u32 tmp=~dlow;
  register u32 res;
  if(tmp) {
    asm volatile ("bsf %1, %0" : "=r"(res) : "rm"(tmp));
    return res;
  } else {
    tmp=~dhigh;
    asm volatile ("bsf %1, %0" : "=r"(res) : "rm"(tmp));
    return res+32;
  }
}

/* Find first set bit in 64-bit word */
inline int bmaskFirstSet64(u32 dlow, u32 dhigh) {
  register u32 res;
  if(dlow) {
    asm volatile ("bsf %1, %0" : "=r"(res) : "rm"(dlow));
    return res;
  } else {
    asm volatile ("bsf %1, %0" : "=r"(res) : "rm"(dhigh));
    return res+32;
  }
}


#ifdef DEBUG_BITS
inline int bmaskCount(u64 data) {
  int count=0;
  while(data) {
    if(data & 1)
      count++;
    data>>=1;
  }
  return count;
}

inline void bmaskCheck(u64 d1, u64 d2) {
  if(bmaskCount(d1)!=bmaskCount(d2) || bmaskCount(d2)<4) {
    printf("bit count changed!");
    bmaskPrint(d1);
    bmaskPrint(d2);
  }
}

# define BIT_CHECK(a,b) bmaskCheck(a,b)
#else
# define BIT_CHECK(a,b)
#endif

// Masks for coord A = 0
#define X0 (XYZMASK(0,0,0) | XYZMASK(0,1,0) | XYZMASK(0,2,0) | XYZMASK(0,3,0) | \
            XYZMASK(0,0,1) | XYZMASK(0,1,1) | XYZMASK(0,2,1) | XYZMASK(0,3,1) | \
            XYZMASK(0,0,2) | XYZMASK(0,1,2) | XYZMASK(0,2,2) | XYZMASK(0,3,2) | \
            XYZMASK(0,0,3) | XYZMASK(0,1,3) | XYZMASK(0,2,3) | XYZMASK(0,3,3) | \
            XYZMASK(0,0,4) | XYZMASK(0,1,4) | XYZMASK(0,2,4) | XYZMASK(0,3,4) )

#define Y0 (XYZMASK(0,0,0) | XYZMASK(1,0,0) | XYZMASK(2,0,0) | \
            XYZMASK(0,0,1) | XYZMASK(1,0,1) | XYZMASK(2,0,1) | \
            XYZMASK(0,0,2) | XYZMASK(1,0,2) | XYZMASK(2,0,2) | \
            XYZMASK(0,0,3) | XYZMASK(1,0,3) | XYZMASK(2,0,3) | \
            XYZMASK(0,0,4) | XYZMASK(1,0,4) | XYZMASK(2,0,4) )

#define Z0 (XYZMASK(0,0,0) | XYZMASK(1,0,0) | XYZMASK(2,0,0) | \
            XYZMASK(0,1,0) | XYZMASK(1,1,0) | XYZMASK(2,1,0) | \
            XYZMASK(0,2,0) | XYZMASK(1,2,0) | XYZMASK(2,2,0) | \
            XYZMASK(0,3,0) | XYZMASK(1,3,0) | XYZMASK(2,3,0) )

// masks to detect bits blocking a rotation via axis AB
#define XYOUT   ( Y0 << XYZOFFSET(0,3,0))
#define XZOUT   ((Z0 << XYZOFFSET(0,0,3)) | (Z0 << XYZOFFSET(0,0,4)))
#define YZOUT   ( Z0 << XYZOFFSET(0,0,4))

// define a rotation function
#define ROTATE_DEFINE(name,mx,my,mz) \
inline u64 name(u64 data) {\
  u64 newData=0;\
  int x,y,z;\
  for(z=0; z<DZ; z++)\
    for(y=0; y<DY; y++)\
      for(x=0; x<DX; x++)\
        if(data & XYZMASK(x,y,z))\
          newData |= XYZMASK(mx,my,mz);\
  BIT_CHECK(data,newData);\
  return newData;\
}

ROTATE_DEFINE(bmaskRotate180XY,DX-1-x,DY-1-y,z)
ROTATE_DEFINE(bmaskRotate180XZ,DX-1-x,y,DZ-1-z)
ROTATE_DEFINE(bmaskRotate180YZ,x,DY-1-y,DZ-1-z)
ROTATE_DEFINE(bmaskRotateXY,DY-1-DY2X-y,x,z)
ROTATE_DEFINE(bmaskRotateXZ,DZ-1-DZ2X-z,y,x)
ROTATE_DEFINE(bmaskRotateYZ,x,DZ-1-DZ2Y-z,y)


/****************/
/* Orientations */
/****************/

/** A part orientation */
typedef struct {
  u64 clip;     // clipping bitmask (indexed with offset, 1=allowed)
  u64 blocks;   // blocks in part   (indexed with offset, 1=present)
  int offset;   // offset of 1st present block.
} Orient;

int    numOrients=0;          // number of orientations
Orient orients[ORIENTATIONS]; // the orientations

void orientPrint(const Orient *o) {
  printf("orient (offset %2d)\n",o->offset);
  bmaskPrint(o->blocks);
#ifdef DEBUG_SHOW_CLIP
  bmaskPrint(o->clip);
#endif
}

// compile pattern to sit in leftmoft upmost frontmost corner,
// compile clipping information, offset.
// return true if nondegenerate pattern
Orient *orientNew(u64 blocks) {
  if(!blocks) {
    printf("Aborting compilation of empty pattern!\n");
    return NULL;
  }

#ifdef DEBUG_BITS
  u64 oldBlocks=blocks;
#endif

  // realign pattern into top left outermost corner
  while(!(blocks & Z0))
    blocks >>= DX*DY;
  while(!(blocks & Y0))
    blocks >>= DX;
  while(!(blocks & X0))
    blocks >>= 1;

  BIT_CHECK(blocks,oldBlocks);

  // assume connected pattern to determine size
  // and set clipping mask accordingly.
  u64 clip=(u64)-1;

#define SHIFT_OP(smask,sand,sshift,sdetect) \
{ u64 mask=(smask);\
  u64 tmp=blocks;\
  while(1) {\
    tmp = (tmp & ~(sand)) >> (sshift);\
    if(!(tmp & (sdetect)))\
      break;\
    clip &= ~mask;\
    mask >>= (sshift);\
  }\
}

  SHIFT_OP(Z0 << XYZOFFSET(0,0,DZ-1),0 ,DX*DY,Z0)
  SHIFT_OP(Y0 << XYZOFFSET(0,DY-1,0),Y0,DX   ,Y0)
  SHIFT_OP(X0 << XYZOFFSET(DX-1,0,0),X0,1    ,X0)
#undef SHIFT_OP

  Orient *o=&(orients[numOrients++]);
  o->clip  =clip;
  o->blocks=blocks;
  o->offset=bmaskFirstSet64(0xffffffffu & (u32)blocks,0xffffffffu & (u32) (blocks>>32));

  return o;
}

Orient *orientInit(const char *init[]) {
  u64 blocks=bmaskInit(init);
  return orientNew(blocks);
}

Orient *orientRotateXY(u64 data) {
  if(data & XYOUT)
    return NULL;
  return orientNew(bmaskRotateXY(data));
}

Orient *orientRotateXZ(u64 data) {
  if(data & XZOUT)
    return NULL;
  return orientNew(bmaskRotateXZ(data));
}

Orient *orientRotateYZ(u64 data) {
  if(data & YZOUT)
    return NULL;
  return orientNew(bmaskRotateYZ(data));
}

Orient *orientRotate180XY(u64 data) {
  u64 newData=bmaskRotate180XY(data);
  return orientNew(newData);
}

Orient *orientRotate180XZ(u64 data) {
  u64 newData=bmaskRotate180XZ(data);
  return orientNew(newData);
}

Orient *orientRotate180YZ(u64 data) {
  u64 newData=bmaskRotate180YZ(data);
  return orientNew(newData);
}

/*********/
/* Parts */
/*********/

/** A part */
typedef struct{
  Orient *orients; // initial orientation
  int     count;   // number of orientations
} Part;

int    initialOrient=0;       // initial orientation index
int    numParts=0;            // number of parts (for init)

Part   parts[PARTS];          // the parts
extern int    nextPart[];

void partPrint(const Part *p) {
  printf("part orientations: %2d\n",p->count);
  int i=0;
#ifdef DEBUG_SHOW_ALL_ORIENTS
  for(i=0; i<p->count; i++)
#endif
    orientPrint(&(p->orients[i]));
}

/** Try to insert a part into the list.
    Checks for duplicates, deletes part if previously created.
    Returns the index where inserted / previously found.
 */
void partInsertRotation(Orient *o) {
  int i;
  for(i=initialOrient; i<numOrients-1; i++)
    if(o->blocks==orients[i].blocks) { // check for duplicates
      numOrients--;                    //   found -> delelete o
      return;
    }
}

#define DEFINE_PART_ROTATE(name,nextinvoke,rot90,rot180) \
void name(u64 blocks) {\
  nextinvoke(blocks);\
  int i;\
  for(i=0; i<3; i++) {\
    Orient *o=rot90(blocks);\
    if(!o) {\
      o=rot180(blocks);\
      i=3;\
    }\
    partInsertRotation(o);\
    blocks=o->blocks;\
    nextinvoke(blocks);\
  }\
}

DEFINE_PART_ROTATE(partRotateYZ,(void)      ,orientRotateYZ,orientRotate180YZ)
DEFINE_PART_ROTATE(partRotateXZ,partRotateYZ,orientRotateXZ,orientRotate180XZ)
DEFINE_PART_ROTATE(partRotateXY,partRotateXZ,orientRotateXY,orientRotate180XY)
#undef DEFINE_PART_ROTATE

Part *partNew(const char *init[]) {
  initialOrient=numOrients;

  Orient *o0=orientInit(init);
  partRotateXY(o0->blocks);

  Part *p=&(parts[numParts++]);
  p->orients=o0;
  p->count  =numOrients - initialOrient;  // continuous allocation

  return p;
}

void partsInit() {
  int i;
  for(i=0; i<PARTS; i++) {
    partNew(partConsts[i]);
    nextPart[i]=i;
  }

#ifdef DEBUG_SHOW_PARTS
  for(i=0; i<PARTS; i++)
    partPrint(parts+nextPart[i]);
#endif

  printf("total orientations: %3d (of %d)\n",numOrients,PARTS*24);
}


/************/
/* Prealign */
/************/

// lookup table for transposed orientations
// by target offset and part.
// holds index into trans array
// (until next one.)
u32 transLookup[DX*DY*DZ*PARTS];

#define TRANS 3425  // << DX*DY*DZ*PARTS*PART_ORIENTS

// transposed rotated parts
u64 trans[TRANS];

void prealignInit() {
  int nextIndex=0;

  int offset;
  for(offset=0; offset<DX*DY*DZ; offset++) {
    int part;
    for(part=0; part<PARTS; part++) {
      transLookup[offset*PARTS+part]=nextIndex;
      const Part *p=&(parts[part]);

#ifdef DEBUG_BITS
      const u64  orig=p->orients->blocks;
#endif

      Orient *o  =p->orients;
      const Orient *end=o+p->count;
      for(; o<end; o++) {
        int pos=offset - o->offset;
        if(pos>=0 && pos<DX*DY*DZ && (o->clip & MASK(pos))) {
          u64 transp=o->blocks << pos;

          BIT_CHECK(orig,transp);

          trans[nextIndex++]=transp;
        }
      }
    }
  }

  printf("total translations: %d (of %d)\n",nextIndex,DX*DY*DZ*PARTS*PART_ORIENTS);
}


/***********/
/* Results */
/***********/

u64    timeStart=0;           // starting time of computation
u32    solutions=0;          // number of solutions found

u64 getTimeMs() {
  struct timeval tv;
  gettimeofday(&tv,NULL);
  return (tv.tv_sec*1000)+tv.tv_usec/1000;
}

void resultsInit() {
  timeStart=getTimeMs();
}

void resultsPrint() {
  u64 timeEnd=getTimeMs();
  u64 timeElapsed=timeEnd-timeStart;
  double solutionsPerS=(1000.0*solutions)/timeElapsed;
  printf("time: %llums solutions: %lu (%.2f /s)\n",
         timeElapsed,solutions,solutionsPerS);
}


/***********/
/* Matcher */
/***********/

u64         board=0;
u32         depth=0;
int         firstFree=0;      // index of first free bit in board
int         nextPart[PARTS];  // remaining unused parts

extern void recurseAsm();


/************/
/* Openings */
/************/

#define MAX_OPENINGS 256
#define MAX_SELFSYMM 4

void openingPlay(int dep, const u64 initial) {
  const int symPart=nextPart[PARTS-dep-1];

#ifdef DEBUG_OPENINGS
  if(dep==0)
    printf("opening at dep %d with symmetry part %d\n",dep,symPart);
#endif

  u32 numOpenings=0;
  u64 openings[MAX_OPENINGS];
  u32 numSelfSymm=0;
  u64 selfSymm[MAX_SELFSYMM];

  int offset;
  for(offset=0; offset<DX*DY*DZ; offset++) {
    int lookup=offset*PARTS+symPart;
    int index=transLookup[lookup];
    const int end=transLookup[lookup+1];
    for(;index<end;index++) {
      const u64 blocks=trans[index];
      if(blocks & initial)
        continue;
      const u64 cand=blocks | initial;

      const u64 rotations[4]={cand,
                              bmaskRotate180XY(cand),
                              bmaskRotate180XZ(cand),
                              bmaskRotate180YZ(cand) };
      bool symmetric=false;

      u32 oldOp;
      for(oldOp=0; (oldOp<numOpenings) && !symmetric; oldOp++) {
        int rot;
        for(rot=0; rot<4; rot++)
          if(openings[oldOp]==rotations[rot]) {
            symmetric=true;
            break;
          }
      }

      for(oldOp=0; (oldOp<numSelfSymm) && !symmetric; oldOp++) {
        int rot;
        for(rot=0; rot<4; rot++)
          if(selfSymm[oldOp]==rotations[rot]) {
            symmetric=true;
            break;
          }
      }

      if(symmetric)  // symmetric duplicate, discard
        continue;

      u32 rot;
      for(rot=1; rot<4; rot++)
        if(cand==rotations[rot])
          symmetric=true;

      if(symmetric)  // self-symmetry, must break
        selfSymm[numSelfSymm++]=cand;
      else
        openings[numOpenings++]=cand;
    }
  }
  if(depth==0)
    printf("%lu asymmetric, %lu self-symmetric openings\n",numOpenings,numSelfSymm);

  // play self-symmetric openings
  u32 i;
  for(i=0; i<numSelfSymm; i++) {
#ifdef DEBUG_OPENINGS
    printf("begin self-symmetric opening %lu at dep %d\n",i,dep);
#endif
    openingPlay(dep+1, selfSymm[i]);
#ifdef DEBUG_OPENINGS
    printf("end self-symmetric opening %lu at dep %d\n",i,dep);
    resultsPrint();
#endif
  }

  // play asymmetric openings
  for(i=0; i<numOpenings; i++) {
#ifdef DEBUG_OPENINGS
    if(dep==0)
      printf("asymmetric opening %lu at dep %d\n",i,dep);
#endif

    const u64 move=openings[i];
    firstFree=bmaskFirstClear64(0xffffffffu & (u32)  move,
                                0xffffffffu & (u32) (move>>32));
    depth=dep+1;
    board=move;
    recurseAsm();

//      recurse(dep+1,move);
//    if(dep==0)
//      recurseu1(dep+1,move);
//    else
//      recurseu2(dep+1,move);

#ifdef DEBUG_OPENINGS
    if(dep==0)
      resultsPrint();
#endif
  }
}

/****************/
/* Main program */
/****************/

int main(int argc, char **argv) {
  puts("c't puzzle solver (c) 2003 Markus L. Noga (markus at noga dot de)");

  resultsInit();
  partsInit();
  prealignInit();

  openingPlay(0,(u64) 0);

  resultsPrint();
  return 0;
}

