#define PARTS     12
#define NEIGHBORS 0x100a

#define TMP1       %eax
#define TMP2       %edx
#define BOARD_LOW  %ebx
#define BOARD_HIGH %edi
#define PI         %esi
#define INDEX      %ecx
#define LAST       %ebp

.extern _board          # initial board cfg
.extern _depth          # initial depth
.extern _firstFree      # first free bit
.extern _solutions      # solution count
.extern _nextPart       # part array

.extern _transLookup    # lookup translations by PARTS*firstFree+part
.extern _trans          # individual translations

.data

.text
.global _recurseAsm

_recurseAsm:
  pushal

  movl _board    , BOARD_LOW       # ecx = boardlow
  movl _board+4  , BOARD_HIGH      # edx = boardhigh

  call recursiveSolve
  popal
  ret


.align 16
recursiveSolve:
  movl $PARTS-1,  PI
  subl _depth,   PI        # PI = PARTS-depth-1 = pi

  js   forAllPartsDone

forThisPart:
  movl  $PARTS,TMP1
  mull  _firstFree             # side effect dx=TMP2 clobbered

  movl _nextPart(,PI,4), TMP2  # TMP2 = part
  addl  TMP1 ,TMP2             # TMP2 = firstFree*PARTS + part

  movl _transLookup(,TMP2,4),INDEX        # INDEX = index
  movl (_transLookup+4)(,TMP2,4),LAST     # LAST = last

  cmpl INDEX, LAST
  jz   forAllTranslationsDone

forThisTranslation:
  movl _trans(,INDEX,8), TMP1      # TMP1 = trans low
  testl TMP1,BOARD_LOW             # check collision with board low
  jnz translationDone

  movl (_trans+4)(,INDEX,8), TMP2  # TMP2 = trans high
  testl TMP2,BOARD_HIGH            # check collision with board high
  jnz translationDone

  orl   TMP1,BOARD_LOW            # place part
  orl   TMP2,BOARD_HIGH

  pushl _firstFree
  pushl INDEX
  pushl LAST

  notl  BOARD_HIGH           # compute new firstFree -> INDEX (=ecx)
  bsfl  BOARD_HIGH,TMP2
  notl  BOARD_HIGH
  addl  $32,TMP2
  notl  BOARD_LOW
  bsfl  BOARD_LOW,INDEX
  #notl  BOARD_LOW          # cancels out with next one
  cmovz TMP2,INDEX
  movl INDEX,_firstFree

  movl $NEIGHBORS,TMP1        # NEIGHBORS << firstFree
  xorl TMP2,TMP2
  xorl LAST,LAST
  shld %cl,TMP1,TMP2
  shll %cl,TMP1
  test $0x20,%cl
  cmovnz TMP1,TMP2
  cmovnz LAST,TMP1

  #notl  BOARD_LOW           # if((~board) & (NEIGHBORS << firstFree)) descend
  testl TMP1,BOARD_LOW
  notl  BOARD_LOW
  jnz   freeNeighbors
  notl  BOARD_HIGH
  testl TMP2,BOARD_HIGH
  notl  BOARD_HIGH
  jz    restore

freeNeighbors:
  movl _depth,TMP2           # depth++
  incl TMP2                  # depth is in TMP2, but still unstored

  cmpl $PARTS,TMP2            # solution found?
  jnz   descend
  incl _solutions
  jmp restore

descend:
  movl TMP2,_depth

  neg  TMP2                  # nextPart[pi]=nextPart[PARTS-1-depth], depth is decremented already!
  movl (_nextPart+4*PARTS)(,TMP2,4), TMP1
  pushl     _nextPart(,PI,4)   # backup current part (nextPart[pi])
  movl TMP1,_nextPart(,PI,4)

  pushl PI

  testl $0x20, INDEX    # call optimized routine if remainder is 32 bit
  jnz  rs32
  call recursiveSolve
  jmp rsDone
rs32:
  call recursiveSolve32
rsDone:

  popl PI

  popl _nextPart(,PI,4)      # nextPart[pi]=part#

  decl _depth                # depth--

restore:
  popl LAST
  popl INDEX

  xorl _trans(,INDEX,8),     BOARD_LOW    # remove piece
  xorl (_trans+4)(,INDEX,8), BOARD_HIGH

  popl _firstFree

translationDone:
  incl INDEX
  cmpl INDEX, LAST
  jnz  forThisTranslation

forAllTranslationsDone:
  decl PI              # PI = pi
  jns  forThisPart

forAllPartsDone:
  ret


################################## top 32 bits only
##################################

.align 16
recursiveSolve32:
  movl $PARTS-1,  PI
  subl _depth,   PI        # PI = PARTS-depth-1 = pi

  js   forAllPartsDone32

forThisPart32:
  movl  $PARTS,TMP1
  mull  _firstFree             # side effect dx=TMP2 clobbered

  movl _nextPart(,PI,4), TMP2  # TMP2 = part
  addl  TMP1 ,TMP2             # TMP2 = firstFree*PARTS + part

  movl _transLookup(,TMP2,4),INDEX        # INDEX = index
  movl (_transLookup+4)(,TMP2,4),LAST     # LAST = last

  cmpl INDEX, LAST
  jz   forAllTranslationsDone32

forThisTranslation32:
  movl (_trans+4)(,INDEX,8), TMP2  # TMP2 = trans high
  testl TMP2,BOARD_HIGH            # check collision with board high
  jnz translationDone32

  orl   TMP2,BOARD_HIGH             # place part

  pushl _firstFree
  pushl INDEX

  notl  BOARD_HIGH           # compute new firstFree -> INDEX (=ecx)
  bsfl  BOARD_HIGH,INDEX
  #notl  BOARD_HIGH      # cancels out with next one
  addl  $32,INDEX
  movl INDEX,_firstFree

  movl $NEIGHBORS,TMP2        # NEIGHBORS << firstFree
  shll %cl,TMP2
  #notl  BOARD_HIGH          # if((~board) & (NEIGHBORS << firstFree)) descend
  testl TMP2,BOARD_HIGH
  notl  BOARD_HIGH
  jz    restore32

freeNeighbors32:
  movl _depth,TMP2           # depth++
  incl TMP2                  # depth is in TMP2, but still unstored

  cmpl $PARTS,TMP2            # solution found?
  jnz   descend32
  incl _solutions
  jmp restore32

descend32:                     # depth is in TMP2, but still unstored
  movl TMP2,_depth

  neg  TMP2                  # nextPart[pi]=nextPart[PARTS-1-depth], depth is decremented already!
  movl (_nextPart+4*PARTS)(,TMP2,4), TMP1
  pushl     _nextPart(,PI,4)   # backup current part (nextPart[pi])
  movl TMP1,_nextPart(,PI,4)

  pushl LAST
  pushl PI
  call recursiveSolve32
  popl PI
  popl LAST

  popl _nextPart(,PI,4)      # nextPart[pi]=part#

  decl _depth                # depth--

restore32:
  popl INDEX

  xorl (_trans+4)(,INDEX,8), BOARD_HIGH   # remove piece

  popl _firstFree

translationDone32:
  incl INDEX
  cmpl INDEX, LAST
  jnz  forThisTranslation32

forAllTranslationsDone32:
  decl PI              # PI = pi
  jns  forThisPart32

forAllPartsDone32:
  ret

.end
